///////////////////////////////////////////////////////////////////////
// WMV plugin for VirtualDub 1.8.1+
// Copyright (C) 2007-2012 fccHandler
// Copyright (C) 1998-2012 Avery Lee
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
///////////////////////////////////////////////////////////////////////

// Main Plugin interfaces

#if defined(_MSC_VER) && (_MSC_VER >= 1400)
	#define _CRT_SECURE_NO_WARNINGS
#endif

#include <windows.h>
#include <commctrl.h>
#include <vfw.h>
#include <stdio.h>

#include "vd2\plugin\vdinputdriver.h"
#include "Unknown.h"
#include "WMVTypes.h"

#include "WinFile.h"
#include "DataVector.h"
#include "ASFStreamNode.h"
#include "InputFileWMV.h"
#include "VideoSourceWMV.h"
#include "AudioSourceWMV.h"
#include "VideoCodec.h"

#include "Resource.h"

#include <crtdbg.h>

extern "C" const VDXPluginInfo *const * VDXAPIENTRY VDGetPluginInfo();

// Patch the Windows Media Audio format header?
// (I'm not sure why this is needed, but it tends
// to not find any WMA codecs without this patch.)

#define PATCH_WMA


static HMODULE g_hModule;

BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
	if (fdwReason == DLL_PROCESS_ATTACH)
		g_hModule = (HMODULE)hinstDLL;

	return TRUE;
}


///////////////////////////////////////////////////////////////////////
//
//						InputFileWMV
//
///////////////////////////////////////////////////////////////////////

InputFileWMV::InputFileWMV(const VDXInputDriverContext& context, uint32 flags)
	: mContext(context)
	, uiFlags (flags)
{
	int i;

	dwPreroll            = 0;
	pktbuf               = NULL;
	pktsize              = 0;
	num_streams          = 0;
	num_audio_streams    = 0;
	languages            = NULL;
	language_size        = 0;

	for (i = 0; i < MAX_STREAMS; ++i)
	{
		streams[i] = NULL;
	}
}


InputFileWMV::~InputFileWMV()
{
	int i;
	for (i = num_streams - 1; i >= 0; --i)
	{
		delete streams[i];
	}
	delete[] pktbuf;
	delete[] languages;
}


char *InputFileWMV::GetLanguageA(int index)
{
	const unsigned char *lang;
	int remains;
	char *ret;

	ret     = NULL;
	lang    = languages;
	remains = language_size;

	if (lang != NULL && remains >= sizeof(WORD))
	{
		WORD Language_ID_Records_Count = *(WORD *)lang;
		remains -= sizeof(WORD);
		lang += sizeof(WORD);

		while (Language_ID_Records_Count != 0)
		{
			BYTE Language_ID_Length;

			if (remains < sizeof(BYTE)) break;

			Language_ID_Length = *lang++;
			--remains;

			if (index == 0)
			{
				int len;
				Language_ID_Length /= sizeof(WCHAR);

				len = WideCharToMultiByte(CP_ACP, 0, (LPCWSTR)lang, Language_ID_Length, NULL, 0, NULL, NULL);
				if (len > 0)
				{
					ret = new char[len + 2];
					if (ret != NULL)
					{
						WideCharToMultiByte(CP_ACP, 0, (LPCWSTR)lang, Language_ID_Length, ret, len + 1, NULL, NULL);
					}
				}
				break;
			}

			lang    += Language_ID_Length;
			remains -= Language_ID_Length;

			--Language_ID_Records_Count;
			--index;
		}
	}

	return ret;
}


// Depending on the value of "flags", we need to read either
// a one, two, or four byte operand from the data. The return
// value tells the number of bytes read.
// (Ugly hack, I know...)

int InputFileWMV::Get_VLC(const void *data, DWORD *VLC, int flags) const
{
	if (data)
	{
		switch (flags & 3)
		{
			case 0x01:
				if (VLC) *VLC = *(BYTE *)data;
				return 1;

			case 0x02:
				if (VLC) *VLC = *(WORD *)data;
				return 2;

			case 0x03:
				if (VLC) *VLC = *(DWORD *)data;
				return 4;
		}
	}

	if (VLC) *VLC = 0;

	return 0;
}


long InputFileWMV::ReadData(void *lpBuffer, DWORD pktnum,
	int offset, long bytes, ASFStreamNode *pasn)
{
	BYTE current_obj;
	int streamno;
	long fragment_point = 0;

	char *lpData = (char *)lpBuffer;

	streamno = pasn->stream_number;
	current_obj = pasn->sample_list[pktnum].obj_number;

	if (bytes <= 0) return 0;
	if (pktnum >= pasn->dwActualLength) return 0;

	mFile.Seek(pasn->sample_list[pktnum].pos);

	while (bytes > 0 && mFile.Pos() < i64MovieLimit)
	{
		int pktleft;
		const unsigned char *pp = pktbuf;
		BYTE Length_Type_Flags;
		BYTE Property_Flags;
		BYTE Payload_Flags;
		BYTE Number_Of_Payloads;
		DWORD Padding_Length = 0;

		mFile.Read(pktbuf, pktsize);

		Length_Type_Flags = *pp++;	// Error Correction Flags?
							// 1....... EC Present (1)
							// .xx..... EC Length Type
							// ...x.... Opaque Data Present
							// ....xxxx EC Data Length

		if (Length_Type_Flags & 0x80)
		{
			int EC_Length_Type = (Length_Type_Flags >> 5) & 3;
			int EC_Data_Length = Length_Type_Flags & 0x0F;

			if (EC_Length_Type == 0) pp += EC_Data_Length;
			Length_Type_Flags = *pp++;
		}

		Property_Flags = *pp++;

		pp += Get_VLC(pp, NULL, Length_Type_Flags >> 1);
		pp += Get_VLC(pp, NULL, Length_Type_Flags >> 5);
		pp += Get_VLC(pp, &Padding_Length, Length_Type_Flags >> 3);

		pp += 6;	// Skip Send Time and Duration

		if (Length_Type_Flags & 1)
		{
			// Multiple payloads
			Payload_Flags = *pp++;
			Number_Of_Payloads = (BYTE)(Payload_Flags & 0x3F);
		}
		else
		{
			// Single payload
			Number_Of_Payloads = 1;
		}

		// At this point we're at the beginning of the payload(s)

		pktleft = (pktsize - (int)(pp - pktbuf)) - Padding_Length;

		// parse down subpackets

		while (Number_Of_Payloads--)
		{
			DWORD Offset_Into_Media_Object;
			DWORD Stream_Number;
			DWORD Media_Object_Number;
			DWORD Payload_Length;
			DWORD Replicated_Data_Length;
			const unsigned char *ppData;
			int i;

			i = Get_VLC(pp, &Stream_Number, Property_Flags >> 6);
			pp += i; pktleft -= i;

			i = Get_VLC(pp, &Media_Object_Number, Property_Flags >> 4);
			pp += i; pktleft -= i;

			i = Get_VLC(pp, &Offset_Into_Media_Object, Property_Flags >> 2);
			pp += i; pktleft -= i;

			i = Get_VLC(pp, &Replicated_Data_Length, Property_Flags);
			pp += i; pktleft -= i;

			Stream_Number &= 0x7F;

			if (Replicated_Data_Length == 1)
			{

				// Compressed payload(s), either Multiple or Single
				// (This used to be referred to as "grouping")

				// All payloads contain sub-payloads from a single stream,
				// and each sub-payload is a complete Media Object

				Offset_Into_Media_Object = 0;

				++pp;	// Skip Presentation Time Delta
				--pktleft;

				if (Length_Type_Flags & 1)
				{
					i = Get_VLC(pp, &Payload_Length, Payload_Flags >> 6);
					pp += i; pktleft -= i;
				}
				else
				{
					Payload_Length = pktleft;
				}

				ppData = pp;

				pp += Payload_Length;
				pktleft -= Payload_Length;

				if ((int)Stream_Number == streamno)
				{
					BYTE Sub_Payload_Data_Length;

					// Since this is a grouped set, if we are still in the middle
					// of fragment assembly we've missed a fragment somewhere.

					if (fragment_point)
					{
						mContext.mpCallbacks->SetError("ASF: Missing fragment (fragpt=%ld, current_obj=%i, pkt=%I64i)", fragment_point, current_obj, mFile.Pos() - pktsize);
						bytes = 0;		// forced exit
						break;
					}

					while (ppData < pp && bytes > 0)
					{
						Sub_Payload_Data_Length = *ppData++;

						if ((signed char)(Media_Object_Number - current_obj) > 0)
						{
							mContext.mpCallbacks->SetError("ASF: Missing object (Media_Object_Number=%i, current_obj=%i)", Media_Object_Number, current_obj);
							bytes = 0;		// forced exit
							break;
						}

						if ((BYTE)Media_Object_Number == current_obj)
						{
							if (offset >= Sub_Payload_Data_Length)
							{
								// skip this object
								offset -= Sub_Payload_Data_Length;
							}
							else
							{
								// fetch the object
								int tc = Sub_Payload_Data_Length - offset;
								if (tc > bytes) tc = bytes;

								memcpy(lpData, ppData + offset, tc);
								lpData += tc;
								bytes -= tc;
								offset = 0;
							}

							// Finished with this object, goto next object
							++pktnum;
							if (pktnum >= pasn->dwActualLength)
							{
								bytes = 0;	// forced exit
								break;
							}
							current_obj = pasn->sample_list[pktnum].obj_number;
						}

						Media_Object_Number = (Media_Object_Number + 1) & 0xFF;
						ppData += Sub_Payload_Data_Length;
					}
				}
			}
			else
			{
				// We already know the size and pts of the object we are
				// looking for, therefore we can skip the replicated data.

				pp += Replicated_Data_Length;
				pktleft -= Replicated_Data_Length;

				if (Length_Type_Flags & 1)
				{
					// Multiple payloads, uncompressed

					// Payload Length (supposed to be a WORD)
					i = Get_VLC(pp, &Payload_Length, Payload_Flags >> 6);
					pp += i; pktleft -= i;
				}
				else
				{
					// Single payload, uncompressed
					Payload_Length = pktleft;
				}

				ppData = pp;

				pp += Payload_Length;
				pktleft -= Payload_Length;

				// Is it what we want?

				if ((int)Stream_Number == streamno)
				{
					// A packet might hold objects earlier than the ones we want, so we
					// can ignore those.  But if we hit one later than the one we
					// want, we've gone too far.  Gotta watch out for the FF->00 rollover.

					// Watch out, #'s may not be sequential; EVA2.ASF is missing the 0x65
					// sequence at 0x39796-0x39f16...

					if ((signed char)(Media_Object_Number - current_obj) > 0)
					{
						mContext.mpCallbacks->SetError("ASF: Missing object (Media_Object_Number=%i, current_obj=%i, pktpos=%I64i)", Media_Object_Number, current_obj, mFile.Pos() - pktsize);
						bytes = 0;		// forced exit
						break;
					}

					if ((BYTE)Media_Object_Number == current_obj)
					{
						long Size_Of_Media_Object;

						// If we are reassembling fragments, make sure we haven't missed one.
						if (Offset_Into_Media_Object > (DWORD)fragment_point)
						{
							mContext.mpCallbacks->SetError("ASF: fragment reassembly error (fragpt=%ld, should be %ld)", Offset_Into_Media_Object, fragment_point);
							bytes = 0;		// forced exit
							break;
						}

						// Update fragment point
						fragment_point = Offset_Into_Media_Object + Payload_Length;

						if (offset < fragment_point)
						{
							// We need this fragment
							int tc;
							int frag_off = Offset_Into_Media_Object - offset;

							if (frag_off >= 0)
							{
								// We want the whole fragment
								tc = bytes - frag_off;
								if (tc > 0)
								{
									if ((DWORD)tc > Payload_Length) tc = Payload_Length;
									memcpy(lpData + frag_off, ppData, tc);
								}
							}
							else
							{
								// We just want a latter portion
								tc = Payload_Length + frag_off;
								if (tc > 0)
								{
									if (tc > bytes) tc = bytes;
									memcpy(lpData, ppData - frag_off, tc);
								}
							}
						}

						Size_Of_Media_Object = pasn->sample_list[pktnum].size;
						if (fragment_point >= Size_Of_Media_Object)
						{
							// Finished with this object
							lpData += Size_Of_Media_Object;
							offset -= Size_Of_Media_Object;
							bytes  -= Size_Of_Media_Object;

							fragment_point = 0;

							if (offset < 0) offset = 0;

							++pktnum;
							if (pktnum >= pasn->dwActualLength)
							{
								bytes = 0;	// forced exit
								break;
							}
							current_obj = pasn->sample_list[pktnum].obj_number;
						}
					}
				}
			}

			if (bytes <= 0) break;
		}
	}

	return (long)(lpData - (char *)lpBuffer);
}


INT_PTR CALLBACK InputFileWMV::ParseDialogProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	InputFileWMV *thisPtr = (InputFileWMV *)GetWindowLongPtr(hDlg, DWLP_USER);

	switch(uMsg) {

	case WM_INITDIALOG:
	{
		char buf[64];

		SetWindowLongPtrA(hDlg, DWLP_USER, lParam);
		thisPtr = (InputFileWMV *)lParam;

		uint32 ver = (*VDGetPluginInfo())->mVersion;

		_snprintf(buf, sizeof(buf), "WMV Plugin v%i.%i by fccHandler",
			(ver >> 24) & 0xFF, (ver >> 16) & 0xFF);

		buf[sizeof(buf) - 1] = '\0';
		SetWindowTextA(hDlg, buf);

		SendMessageA(GetDlgItem(hDlg, IDC_PROGRESS1), PBM_SETRANGE, 0, MAKELPARAM(0, 16384));
		SetTimer(hDlg, 1, 250, (TIMERPROC)NULL);
		ShowWindow(hDlg, SW_SHOW);

		return TRUE;
	}

	case WM_TIMER:
	{
		char buf[64];
		__int64 file_cpos = thisPtr->mFile.Pos();
		__int64 file_size = thisPtr->mFile.Size();

		SendMessageA(GetDlgItem(hDlg, IDC_PROGRESS1), PBM_SETPOS,
			(WPARAM)((file_cpos * 16384I64) / file_size), 0);

		_snprintf(buf, sizeof(buf),
			"%I64iK of %I64iK", file_cpos / 1024,
			(file_size + 512) / 1024);
			
		buf[sizeof(buf) - 1] = '\0';
		SendDlgItemMessageA(hDlg, IDC_CURRENT_VALUE, WM_SETTEXT, 0, (LPARAM)buf);

		return TRUE;
	}

	case WM_COMMAND:
		if (LOWORD(wParam) == IDCANCEL) thisPtr->fAbort = true;

		return TRUE;
	}

	return FALSE;
}


bool InputFileWMV::guiDlgMessageLoop(HWND hDlg)
{
	MSG msg;

	while(PeekMessageA(&msg, NULL, 0, 0, PM_REMOVE))
	{
		if (msg.message == WM_QUIT) return false;

		if (!hDlg || !IsWindow(hDlg) || !IsDialogMessageA(hDlg, &msg))
		{
			TranslateMessage(&msg);
			DispatchMessageA(&msg);
		}
	}

	return true;
}


BOOL CALLBACK InputFileWMV::FindContextWindow(HWND hwnd, LPARAM lParam)
{
	if (GetParent(hwnd) == NULL)
	{
		*(HWND *)lParam = hwnd;
		return FALSE;
	}
	return TRUE;
}


void InputFileWMV::Parse_ASF_Data_Object(__int64 len)
{
	ASFSampleInfo asi_list[128];
	ASF_Data_Object hdr;
	HWND hwndProgress;
	HWND hwndParent;

	for (int i = 0; i < 128; ++i)
	{
		asi_list[i].flags = 0;
	}

	hwndParent   = NULL;
	hwndProgress = NULL;
	fAbort       = false;

	if ((uiFlags & IVDXInputFileDriver::kFlagQuiet) == 0)
	{
		// Get a handle to the main VirtualDub window
		EnumThreadWindows(GetCurrentThreadId(), FindContextWindow, (LPARAM)&hwndParent);
		if (hwndParent == NULL) hwndParent = GetActiveWindow();
		if (hwndParent != NULL) EnableWindow(hwndParent, FALSE);

		// Pop up parse dialog

		hwndProgress = CreateDialogParamA(g_hModule,
			MAKEINTRESOURCEA(IDD_PROGRESS), hwndParent,
			ParseDialogProc, (LPARAM)this);
	}

	mFile.Read(&hdr, sizeof(hdr));
	len -= sizeof(hdr);

	// Now we are right at the beginning of the packets

	while (len >= pktsize)
	{
		const unsigned char *pp = pktbuf;

		BYTE	Length_Type_Flags;
		BYTE	Property_Flags;
		BYTE	Payload_Flags;
		BYTE	Number_Of_Payloads;
		DWORD	Packet_Length;
		DWORD	Padding_Length;

		int pktleft;

		if (hwndProgress != NULL)
		{
			// Update the progress dialog
			guiDlgMessageLoop(hwndProgress);
			if (fAbort) break;
		}

		// The following reads the whole packet
		mFile.Read(pktbuf, pktsize);
		len -= pktsize;

		Length_Type_Flags = *pp++;	// Error Correction Flags?
							// 1....... EC Present (1)
							// .xx..... EC Length Type
							// ...x.... Opaque Data Present
							// ....xxxx EC Data Length

		if (Length_Type_Flags & 0x80)
		{
			int EC_Length_Type = (Length_Type_Flags >> 5) & 3;
			int EC_Data_Length = Length_Type_Flags & 0x0f;

			if (EC_Length_Type == 0) pp += EC_Data_Length;
			Length_Type_Flags = *pp++;
		}

		Property_Flags = *pp++;

		pp += Get_VLC(pp, NULL, Length_Type_Flags >> 1);
		pp += Get_VLC(pp, &Packet_Length, Length_Type_Flags >> 5);
		pp += Get_VLC(pp, &Padding_Length, Length_Type_Flags >> 3);

		pp += 6;	// skip Send_Time and Duration

		// Note that "Duration" is the duration of
		// the packet; it's unrelated to the duration
		// of the media objects inside the packet!

		if (Length_Type_Flags & 1)
		{
			// Multiple payloads
			Payload_Flags = *pp++;

			Number_Of_Payloads = Payload_Flags & 0x3F;
		}
		else
		{
			// Single payload
			Number_Of_Payloads = 1;
		}

		// At this point we're at the
		// beginning of the payload(s)

		pktleft = (pktsize - (int)(pp - pktbuf)) - Padding_Length;

		// parse down segments/subpackets/whatever...

		// Whether Multiple or Single, right now
		// we are at the "Stream_Number" field

		while (Number_Of_Payloads--)
		{
			DWORD Media_Object_Number;
			DWORD Offset_Into_Media_Object;
			bool Key_Frame_Bit;
			DWORD Stream_Number;
			DWORD Payload_Length;
			DWORD Replicated_Data_Length;
			DWORD Presentation_Time;
			int i;

			i = Get_VLC(pp, &Stream_Number, Property_Flags >> 6);
			pp += i; pktleft -= i;

			i = Get_VLC(pp, &Media_Object_Number, Property_Flags >> 4);
			pp += i; pktleft -= i;

			// The following is Offset Into Media Object for uncompressed
			// payloads, but it's Presentation Time for compressed payloads:

			i = Get_VLC(pp, &Offset_Into_Media_Object, Property_Flags >> 2);
			pp += i; pktleft -= i;

			i = Get_VLC(pp, &Replicated_Data_Length, Property_Flags);
			pp += i; pktleft -= i;

			Key_Frame_Bit = !!(Stream_Number & 0x80);
			Stream_Number &= 0x7F;

			if (Replicated_Data_Length == 1)
			{
				// Compressed payload(s), either Multiple or Single
				// (This used to be referred to as "grouping")

				// All payloads contain sub-payloads from a single stream,
				// and each sub-payload is a complete Media Object

				BYTE Presentation_Time_Delta;
				const unsigned char *pplimit;

				Presentation_Time = Offset_Into_Media_Object; // - dwPreroll;
				Offset_Into_Media_Object = 0;

				Presentation_Time_Delta = *pp++;
				--pktleft;

				if (Length_Type_Flags & 1)
				{
					// Multiple payloads

					// Payload Length (supposed to be a WORD)
					i = Get_VLC(pp, &Payload_Length, Payload_Flags >> 6);
					pp += i; pktleft -= i;
				}
				else
				{
					// Single payload

					Payload_Length = pktleft;
				}

				pktleft -= Payload_Length;
				pplimit = pp + Payload_Length;

				while (pp < pplimit)
				{
					BYTE Sub_Payload_Data_Length = *pp++;

					//pp += Sub_Payload_Data_Length;

					// Add sample to the appropriate stream
					for (i = 0; i < num_streams; ++i)
					{
						if (streams[i]->stream_number == (int)Stream_Number)
						{
							ASFSampleInfo *asi = &(asi_list[Stream_Number]);

							asi->pos = mFile.Pos() - pktsize;
							asi->size = Sub_Payload_Data_Length;
							asi->pres_time = Presentation_Time;
							asi->obj_number = (BYTE)Media_Object_Number;
							asi->flags = Key_Frame_Bit? ASI_FLAG_KEYFRAME: 0;

							streams[i]->dv->Add(asi);
							break;
						}
					}
					pp += Sub_Payload_Data_Length;

					Presentation_Time += Presentation_Time_Delta;
					Media_Object_Number = (Media_Object_Number + 1) & 0xFF;
				}
			}
			else if (Replicated_Data_Length == 0 || Replicated_Data_Length >= 8)
			{
				DWORD Size_Of_Media_Object;

				ASFSampleInfo *asi = &(asi_list[Stream_Number]);

				// Uncompressed payload(s), either Multiple or Single
				// (Payloads need not contain an entire Media Object)

				if (Replicated_Data_Length >= 8)
				{
					asi->size = *(DWORD *)(pp);
					asi->pres_time = *(DWORD *)(pp+4);

					pp += Replicated_Data_Length;
					pktleft -= Replicated_Data_Length;
				}

				Size_Of_Media_Object = asi->size;
				Presentation_Time = asi->pres_time;

				if (Length_Type_Flags & 0x01)
				{
					// Multiple payloads, uncompressed

					// Payload Length (supposed to be a WORD)
					i = Get_VLC(pp, &Payload_Length, Payload_Flags >> 6);
					pp += i; pktleft -= i;
				}
				else
				{
					// Single payload, uncompressed
					Payload_Length = pktleft;
				}

				//pp += Payload_Length;
				pktleft -= Payload_Length;

				// Add the object to the appropriate stream.  For fragmented
				// objects, only add it once, and only if it's complete.  Keep
				// in mind that a stream might start with an incomplete object
				// (if recording of a live stream began during a fragmented
				// object after the first fragment had been transmitted).

				if (Offset_Into_Media_Object == 0)
				{
					asi->pos = mFile.Pos() - pktsize;
					asi->size = Size_Of_Media_Object;
					asi->pres_time = Presentation_Time;
					asi->obj_number = (BYTE)Media_Object_Number;
					asi->flags = ASI_FLAG_SEEN_FIRST_FRAGMENT;
					asi->flags |= Key_Frame_Bit? ASI_FLAG_KEYFRAME: 0;
				}
				pp += Payload_Length;

				if (((Offset_Into_Media_Object + Payload_Length) >= Size_Of_Media_Object) &&
					(asi->flags & ASI_FLAG_SEEN_FIRST_FRAGMENT))
				{
					asi->flags -= ASI_FLAG_SEEN_FIRST_FRAGMENT;

					// Add sample to the appropriate stream
					for (i = 0; i < num_streams; ++i)
					{
						if (streams[i]->stream_number == (int)Stream_Number)
						{
							ASFStreamNode *sn = streams[i];

							// Check for AVCDecoderConfigurationRecord
							if ( sn->fccType == streamtypeVIDEO
							  && sn->pAVCDCR == NULL
							  && sn->pFormat != NULL
							  && ((BITMAPINFOHEADER *)sn->pFormat)->biCompression == '462H')
							{
								sn->pAVCDCR = new char[asi->size];
								if (sn->pAVCDCR != NULL)
								{
									sn->iAVCDCRLen = asi->size;
									memcpy(sn->pAVCDCR, pp - Payload_Length, sn->iAVCDCRLen);
								}
							}
							else
							{
								streams[i]->dv->Add(asi);
							}
							break;
						}
					}
				}
			}
			else
			{
				// Hmm, I haven't seen this fail yet:
				mContext.mpCallbacks->SetError("ASF: Unhandled Replicated_Data_Length %i in packet at %I64i", Replicated_Data_Length, mFile.Pos() - pktsize);

				fAbort = true;
				break;
			}
		}

		if (fAbort) break;
	}

	if (hwndParent != NULL) EnableWindow(hwndParent, TRUE);
	if (hwndProgress != NULL) DestroyWindow(hwndProgress);
}


bool InputFileWMV::Parse_ASF_Stream_Properties_Object(__int64 limit)
{
	ASF_Stream_Properties_Object hdr;
	ASFStreamNode *pasn;

	unsigned char *Type_Specific_Data = NULL;
	unsigned char *Error_Correction_Data = NULL;

	bool ret = false;

	mFile.Read(&hdr, sizeof(hdr));

	if (hdr.Type_Specific_Data_Length > 0)
	{
		Type_Specific_Data = new unsigned char[hdr.Type_Specific_Data_Length];
		if (Type_Specific_Data == NULL)
		{
			mContext.mpCallbacks->SetErrorOutOfMemory();
			goto Abort;
		}
		mFile.Read(Type_Specific_Data, hdr.Type_Specific_Data_Length);
	}

	if (hdr.Error_Correction_Data_Length > 0)
	{
		Error_Correction_Data = new unsigned char[hdr.Error_Correction_Data_Length];
		if (Error_Correction_Data == NULL)
		{
			mContext.mpCallbacks->SetErrorOutOfMemory();
			goto Abort;
		}
		mFile.Read(Error_Correction_Data, hdr.Error_Correction_Data_Length);
	}

	// Is it a new video or audio stream?

	if (hdr.Stream_Type == GUID_ASF_Video_Media ||
	    hdr.Stream_Type == GUID_ASF_Audio_Media)
	{
		int i;

		if (hdr.Flags & 0x8000)
		{
			mContext.mpCallbacks->SetError(
				"ASF: %s Stream is encrypted",
				(hdr.Stream_Type == GUID_ASF_Video_Media)?
				"Video": "Audio"
			);
			goto Abort;
		}

		// Stream already exists?
		for (i = 0; i < num_streams; ++i)
		{
			if (streams[i]->stream_number == (hdr.Flags & 0x7F))
			{
				// Must have been created by the extended properties parser
				if (streams[i]->fccType != 0)
				{
					mContext.mpCallbacks->SetError(
						"ASF: Conflicting properties for stream %i",
						streams[i]->stream_number
					);
					goto Abort;
				}

				pasn = streams[i];
				break;
			}
		}

		if (i == num_streams)
		{
			// Create a new stream node

			if (num_streams >= MAX_STREAMS)
			{
				// Can't add any more streams
				ret = true;
				goto Abort;
			}

			pasn = streams[num_streams] = new ASFStreamNode;
			if (pasn == NULL)
			{
				mContext.mpCallbacks->SetErrorOutOfMemory();
				goto Abort;
			}
			++num_streams;
		}

		pasn->dv = new DataVector(sizeof(ASFSampleInfo));
		if (pasn->dv == NULL)
		{
			--num_streams;
			delete streams[num_streams];
			streams[num_streams] = NULL;
			mContext.mpCallbacks->SetErrorOutOfMemory();
			goto Abort;
		}

		pasn->stream_number = hdr.Flags & 0x7F;

		if (hdr.Stream_Type == GUID_ASF_Video_Media)
		{
			// Video Stream, starts with:
			//	DWORD	Encoded Image Width
			//	DWORD	Encoded Image Height
			//	BYTE	Reserved Flags
			//	WORD	Format Data Size

			const BITMAPINFOHEADER *bmih =
				(BITMAPINFOHEADER *)(Type_Specific_Data + 11);

			pasn->fccType		= streamtypeVIDEO;
			pasn->dwRate		= 15000;
			pasn->dwScale		= 1000;

			pasn->iFormatLen = *((WORD *)(Type_Specific_Data + 9));
			pasn->pFormat = new char[pasn->iFormatLen];

			if (pasn->pFormat == NULL)
			{
				mContext.mpCallbacks->SetErrorOutOfMemory();
				goto Abort;
			}

			memcpy(pasn->pFormat, bmih, pasn->iFormatLen);
			((BITMAPINFOHEADER *)pasn->pFormat)->biSize = pasn->iFormatLen;
		}
		else
		{
			// Audio Stream

			static const char sGuid[] = "1A0F78F0-EC8A-11d2-BBBE-006008320064";

			const WAVEFORMATEX *wfex = (WAVEFORMATEX *)Type_Specific_Data;

			pasn->fccType		= streamtypeAUDIO;
			pasn->dwRate		= wfex->nAvgBytesPerSec;
			pasn->dwScale		= wfex->nBlockAlign;

			// Patch?

			// These GUIDs are mysterious:
			//	1A0F78F0-EC8A-11d2-BBBE-006008320064
			//	F6DC9830-BC79-11d2-A9D0-006097926036

#ifdef PATCH_WMA
			// If it's Microsoft Audio Codec, manually patch the format
			if (wfex->wFormatTag == 0x160)
			{
				// Windows Media Audio v1
			
				pasn->iFormatLen = sizeof(WAVEFORMATEX) + 4 + sizeof(sGuid);
				pasn->pFormat = new char[pasn->iFormatLen];

				if (pasn->pFormat == NULL) {
					mContext.mpCallbacks->SetErrorOutOfMemory();
					goto Abort;
				}

				memcpy(pasn->pFormat, wfex, sizeof(WAVEFORMATEX) + 4);
				memcpy((char *)pasn->pFormat + sizeof(WAVEFORMATEX) + 4, sGuid, sizeof(sGuid));

				((WAVEFORMATEX *)pasn->pFormat)->cbSize = 4 + sizeof(sGuid);
			}
			else if (wfex->wFormatTag == 0x161)
			{
				// Windows Media Audio v2

				pasn->iFormatLen = sizeof(WAVEFORMATEX) + 10 + sizeof(sGuid);
				pasn->pFormat = new char[pasn->iFormatLen];

				if (pasn->pFormat == NULL) {
					mContext.mpCallbacks->SetErrorOutOfMemory();
					goto Abort;
				}

				memcpy(pasn->pFormat, wfex, sizeof(WAVEFORMATEX) + 10);
				memcpy((char *)pasn->pFormat + sizeof(WAVEFORMATEX) + 10, sGuid, sizeof(sGuid));

				((WAVEFORMATEX *)pasn->pFormat)->cbSize = 10 + sizeof(sGuid);
			}
			else
#endif
			{
				pasn->iFormatLen = hdr.Type_Specific_Data_Length;
				pasn->pFormat = new char[pasn->iFormatLen];

				if (pasn->pFormat == NULL) {
					mContext.mpCallbacks->SetErrorOutOfMemory();
					goto Abort;
				}

				memcpy(pasn->pFormat, wfex, pasn->iFormatLen);
			}

			if (hdr.Error_Correction_Type == GUID_ASF_Audio_Spread
				&& hdr.Error_Correction_Data_Length >= 8)
			{
				const unsigned short *ecc =
					(unsigned short *)(Error_Correction_Data + 1);

				// Span: the number of packets over
				// which audio will be spread (usually 1)
				pasn->iSpan = Error_Correction_Data[0];

				// Virtual Pack Length / Virtual Chunk Length
				// (This result should always be 1)
				if (ecc[1]) {
					pasn->iAudioScrambleChunk = ecc[0] / ecc[1];
				} else {
					pasn->iAudioScrambleChunk = 1;
				}
			}
		}
	}

	ret = true;

Abort:
	delete [] Error_Correction_Data;
	delete [] Type_Specific_Data;

	return ret;
}


bool InputFileWMV::Parse_ASF_Extended_Stream_Properties_Object(__int64 limit)
{
	ASF_Extended_Stream_Properties_Object hdr;
	int i;

	mFile.Read(&hdr, sizeof(hdr));

	// Skip over stream names
	for (i = 0; i < hdr.Stream_Name_Count; ++i)
	{
		ASF_Stream_Name sn_hdr;

		mFile.Read(&sn_hdr, sizeof(sn_hdr));
		if (sn_hdr.Stream_Name_Length)
		{
			mFile.Seek(mFile.Pos() + sn_hdr.Stream_Name_Length);
		}
	}

	// Skip over extensions
	for (i = 0; i < hdr.Payload_Extension_System_Count; ++i)
	{
		ASF_Payload_Extension_System pes_hdr;

		mFile.Read(&pes_hdr, sizeof(pes_hdr));

		if (pes_hdr.Extension_System_Info_Length)
		{
			mFile.Seek(mFile.Pos() + pes_hdr.Extension_System_Info_Length);
		}
	}

	// If anything remains it should be stream properties
	if ((limit - mFile.Pos()) >= sizeof(ASF_Stream_Properties_Object))
	{
		ASF_Object obj;

		mFile.Read(&obj, sizeof(obj));

		if (!Parse_ASF_Stream_Properties_Object(limit)) return false;

		// Apply language index
		for (i = num_streams - 1; i >= 0; --i)
		{
			if (streams[i]->stream_number == hdr.Stream_Number)
			{
				streams[i]->iLanguageIndex = hdr.Stream_Language_ID_Index;
				break;
			}
		}
	}
	else
	{
		// Create the stream anyway; properties should arrive later.

		if (num_streams < MAX_STREAMS)
		{
			ASFStreamNode *pasn;
			pasn = streams[num_streams] = new ASFStreamNode;

			if (pasn == NULL)
			{
				mContext.mpCallbacks->SetErrorOutOfMemory();
				goto Abort;
			}
			pasn->stream_number  = hdr.Stream_Number;
			pasn->iLanguageIndex = hdr.Stream_Language_ID_Index;
			++num_streams;
		}
	}

	return true;

Abort:
	return false;
}


bool InputFileWMV::Parse_ASF_Header_Extension_Object(__int64 limit)
{
	ASF_Header_Extension_Object hdr;
	ASF_Object obj;
	__int64 curpos;

	bool ret = true;

	mFile.Read(&hdr, sizeof(hdr));
	curpos = mFile.Pos();

	while (curpos < limit)
	{
		mFile.Read(&obj, sizeof(obj));

		if (obj.Object_ID == GUID_ASF_Extended_Stream_Properties_Object)
		{
			ret = Parse_ASF_Extended_Stream_Properties_Object(curpos + obj.Object_Size);
			if (!ret) break;
		}
		else if (obj.Object_ID == GUID_ASF_Language_List_Object && obj.Object_Size >= sizeof(WORD))
		{
			delete[] languages;
			language_size = 0;

			languages = new unsigned char[(size_t)obj.Object_Size];
			if (languages == NULL)
			{
				ret = false;
				break;
			}
			language_size = (int)obj.Object_Size;
			mFile.Read(languages, language_size);
		}

		curpos += obj.Object_Size;
		mFile.Seek(curpos);
	}

	return ret;
}


void InputFileWMV::Parse_ASF_File_Properties_Object(__int64 limit)
{
	ASF_File_Properties_Object hdr;

	if ((limit - mFile.Pos()) >= sizeof(hdr))
	{
		mFile.Read(&hdr, sizeof(hdr));

		dwPreroll = (DWORD)hdr.Preroll;
		pktsize = hdr.Maximum_Data_Packet_Size;
	}
}


bool InputFileWMV::Parse_ASF_Header_Object(__int64 plimit)
{

	// ASF Header Object is the only object which can
	// contain other objects.  It must contain the following
	// to be valid:
	//		ASF File Properties Object
	//		ASF Stream Properties Object

	ASF_Header_Object asfhdr;
	__int64 limit;

	bool ret = true;

	if (mFile.Read(&asfhdr, sizeof(asfhdr)) >= sizeof(asfhdr))
	{
		ASF_Object obj;

		while (asfhdr.Number_Of_Header_Objects-- > 0)
		{
			__int64 pos_start = mFile.Pos();
			if (pos_start >= plimit) break;

			mFile.Read(&obj, sizeof(obj));
			limit = pos_start + obj.Object_Size;

			if (obj.Object_Size < sizeof(obj) || limit > mFile.Size())
			{
				// Invalid object size
				break;
			}

			if (obj.Object_ID == GUID_ASF_File_Properties_Object)
			{
				Parse_ASF_File_Properties_Object(limit);
			}
			else if (obj.Object_ID == GUID_ASF_Stream_Properties_Object)
			{
				ret = Parse_ASF_Stream_Properties_Object(limit);
				if (!ret) break;
			}
			else if (obj.Object_ID == GUID_ASF_Header_Extension_Object)
			{
				ret = Parse_ASF_Header_Extension_Object(limit);
				if (!ret) break;
			}

			mFile.Seek(limit);
		}
	}

	return ret;
}


inline void InputFileWMV::DeriveFrameRate(ASFStreamNode *video)
{
	const ASFSampleInfo *asi = video->sample_list;

	int i, t;
	long len;
	int histo[500];
	double msperframe;
	double min;

	len = video->dwActualLength;

	// Compute histogram of differences in presentation time.

	memset(histo, 0, sizeof histo);

	t = asi->pres_time;		// first PTS

	for (i = 1; i < len; ++i)
	{
		int diff = asi[i].pres_time - t;

		// Ignore 0ms differences... this can be due to grouping.

		if (diff > 0)
		{
			if (diff <= 500)
			{
				++histo[diff - 1];
			}
			t = asi[i].pres_time;	// last valid PTS
		}
	}

	// Pick out the first frame difference that occurs for at least
	// 10% of the video.  Ignore any timings below 10ms (100fps).

	t = (len + 5) / 10;		// threshold

	// Note that "threshold" will be zero if length is less
	// than 5.  We do not want histo[i] to be zero!

	for (i = 10; i < 500; ++i)
	{
		if (histo[i] > t) break;
	}

	// Remember the diff was stored in slot[diff - 1]!

	// Now we have histo[i] frames with diff (i + i) as a base
	// Add any other rates which are +/-2 ms from (i + 1)...

	if (i < 500)
	{
		double diff_cnt = histo[i];
		msperframe = (double)histo[i] * (i + 1);
		
		if (i > 10)
		{
			diff_cnt += histo[i - 1];
			msperframe += (double)histo[i - 1] * i;
			if (i > 11)
			{
				diff_cnt += histo[i - 2];
				msperframe += (double)histo[i - 2] * (i - 1);
			}
		}

		if (i < 499)
		{
			diff_cnt += histo[i + 1];
			msperframe += (double)histo[i + 1] * (i + 2);
			if (i < 498)
			{
				diff_cnt += histo[i + 2];
				msperframe += (double)histo[i + 2] * (i + 3);
			}
		}

		msperframe /= diff_cnt;
	}
	else
	{
		// No rate was used for at least 10% of the movie?
		// In that case, just settle for a wild guess
		if (len > 1)
		{
			msperframe = ((double)asi[len - 1].pres_time
				- (double)asi[0].pres_time) / (double)(len - 1);
		}
		else
		{
			// If length is only 1 frame, just use anything
			msperframe = 40;
		}
	}

	// Let's try to lock the rate to something exact:

	static const VDXFraction rates[12] = {
		{60000, 1000},	// 60		60p
		{60000, 1001},	// 59.94	NTSC double rate
		{30000, 1000},	// 30		30p
		{30000, 1001},	// 29.97	NTSC
		{25000, 1000},	// 25		PAL/SECAM
		{24000, 1000},	// 24		Film
		{24000, 1001},	// 23.976	NTSC film
		{18000, 1000},	// 18		Silent film
		{18000, 1001},	// 17.982	NTSC silent film
		{16000, 1000},	// 16		Silent film
		{16000, 1001},	// 15.984	NTSC silent film
		{15000, 1000}	// 15		Half 30p
	};

	// Find the rate with the least amount
	// of deviation (in ms per frame):

	min = 1.0;

	for (i = 0; i < 12; ++i)
	{
		double diff = msperframe -
			((1000.0 * (double)rates[i].mDenominator)
			/ (double)rates[i].mNumerator);

		if (diff < 0) diff = -diff;

		if (diff < min)
		{
			min = diff;
			t = i;		// closest so far
		}
	}

	// Finally, let's accumulate the deviation for the duration
	// of the movie, and arbitrarily say that we'll allow
	// no more than 10ms of deviation for that whole period.

	// The amount of ms we might lose by locking is:
	// deviation_in_ms_per_frame * fps * duration_in_seconds

	// The following is actually measuring one frame less than
	// the total duration, but heck, we're already being more
	// than nitpicky about this...

	min *= (1000.0 / msperframe) *
		((double)(asi[len - 1].pres_time - asi[0].pres_time) / 1000.0);

	if (min < 10.0)
	{
		// Awesome; lock to this
		video->dwRate = rates[t].mNumerator;
		video->dwScale = rates[t].mDenominator;
	}
	else
	{
		// Measured rate is too far off; don't lock.
		// Instead write it as microseconds per frame
		video->dwRate = 1000000000;
		video->dwScale = (DWORD)(msperframe * 1000000.0 + 0.5);
	}
}


void InputFileWMV::DeriveDuration(ASFStreamNode& asn)
{
	asn.Duration = 1.0;

	if (asn.dwActualLength != 0)
	{
		unsigned long dwFirstPacket = 0;
		unsigned long dwLastPacket  = 0;
		unsigned long dwEarliestPTS = asn.sample_list[0].pres_time;
		unsigned long dwLatestPTS   = dwEarliestPTS;

		for (unsigned long i = 1; i < asn.dwActualLength; ++i)
		{
			if (asn.sample_list[i].pres_time < dwEarliestPTS)
			{
				dwFirstPacket = i;
				dwEarliestPTS = asn.sample_list[i].pres_time;
			}
			if (asn.sample_list[i].pres_time > dwLatestPTS)
			{
				dwLastPacket  = i;
				dwLatestPTS   = asn.sample_list[i].pres_time;
			}
		}

		if (dwLastPacket > dwFirstPacket)
		{
			double msperframe;

			msperframe = (double)(dwLatestPTS  - dwEarliestPTS)
			           / (double)(dwLastPacket - dwFirstPacket);

			asn.Duration = msperframe * 0.001 * asn.dwActualLength;
		}
	}
}


void InputFileWMV::ParseFileRoot()
{
	ASFStreamNode *pasn;
	ASF_Object hdr;
	int s;

	__int64 i64MovieLen;
	__int64 i64MoviePos = 0;

	pktsize = 0;

	// ASF Header Object is mandatory

	mFile.Seek(0);
	if (mFile.Read(&hdr, sizeof(hdr)) >= sizeof(hdr))
	{
		if (hdr.Object_ID == GUID_ASF_Header_Object
			&& hdr.Object_Size >= sizeof(hdr)
			&& hdr.Object_Size < (QWORD)mFile.Size())
		{
			Parse_ASF_Header_Object(hdr.Object_Size);
			mFile.Seek(hdr.Object_Size);
		}
	}

	if (pktsize == 0)
	{
		mContext.mpCallbacks->SetError("ASF Header Object is invalid");
		return;
	}

	if (mFile.Read(&hdr, sizeof(hdr)) >= sizeof(hdr))
	{
		if (hdr.Object_ID == GUID_ASF_Data_Object
			&& hdr.Object_Size >= sizeof(hdr))
		{
			i64MoviePos = mFile.Pos();
			i64MovieLen = hdr.Object_Size - sizeof(hdr);
			if (i64MovieLen > (mFile.Size() - i64MoviePos))
			{
				i64MovieLen = mFile.Size() - i64MoviePos;
			}
		}
	}

	if (i64MoviePos == 0 || i64MovieLen <= 0)
	{
		mContext.mpCallbacks->SetError("ASF Data Object is invalid");
		return;
	}

	i64MovieLimit = i64MoviePos + i64MovieLen;

	// Allocate packet buffer

	pktbuf = new unsigned char[pktsize];
	if (pktbuf == NULL)
	{
		mContext.mpCallbacks->SetErrorOutOfMemory();
		return;
	}

	Parse_ASF_Data_Object(i64MovieLen);

	// Voila, the WMV is parsed!


	// Create stream sample lists

	for (s = 0; s < num_streams; ++s)
	{
		streams[s]->sample_list = (ASFSampleInfo *)streams[s]->dv->MakeArray();
		streams[s]->dwActualLength = streams[s]->dv->Length();
		delete streams[s]->dv;
		streams[s]->dv = NULL;
	}

	// Remove any zero length streams

	s = 0;
	while (s < num_streams)
	{
		if (streams[s]->dwActualLength == 0)
		{
			int i;
			delete streams[s];
			for (i = s + 1; i < num_streams; ++i)
			{
				streams[i - 1] = streams[i];
			}
			--num_streams;
			streams[num_streams] = NULL;
		}
		else ++s;
	}

	// Preprocess some things that all streams have in common.
	for (s = 0; s < num_streams; ++s)
	{
		DWORD i;
		pasn = streams[s];

		pasn->i64TotalBytes = 0;
		pasn->dwMaxSample  = pasn->sample_list[0].size;
		pasn->bHasVSB      = false;

		// Subtract preroll from all time stamps.
		// Determine the largest size of any sample.
		// Get the total number of bytes.
		for (i = 0; i < pasn->dwActualLength; ++i)
		{
			if (pasn->sample_list[i].pres_time >= dwPreroll) {
				pasn->sample_list[i].pres_time -= dwPreroll;
			} else {
				pasn->sample_list[i].pres_time = 0;
			}
			if ((DWORD)pasn->sample_list[i].size != pasn->dwMaxSample)
			{
				pasn->bHasVSB = true;
				if ((DWORD)pasn->sample_list[i].size > pasn->dwMaxSample) {
					pasn->dwMaxSample = pasn->sample_list[i].size;
				}
			}
			pasn->i64TotalBytes += pasn->sample_list[i].size;
		}

		DeriveDuration(*pasn);
	}

	// Get the first video stream found

	pasn = NULL;
	for (s = 0; s < num_streams; ++s)
	{
		if (streams[s]->fccType == streamtypeVIDEO)
		{
			pasn = streams[s];
			break;
		}
	}

	if (pasn != NULL)
	{
		DWORD i;

		DeriveFrameRate(pasn);

		i = pasn->dwActualLength;
		pasn->dwPerceivedLength = i;

		if (i > 1)
		{
			DWORD current;

			// Calculate the perceived length in frames

			pasn->dwPerceivedLength = (DWORD)
				((pasn->Duration * pasn->dwRate) / pasn->dwScale + 0.5);

			// Now generate a mapping from perceived frame to actual frame

			pasn->sample_map = new unsigned long[pasn->dwPerceivedLength];
			if (pasn->sample_map == NULL)
			{
				mContext.mpCallbacks->SetErrorOutOfMemory();
				return;
			}

			current = 0;
			pasn->sample_map[0] = 0;

			for (i = 1; i < pasn->dwPerceivedLength; ++i)
			{
				// Calculate the ideal presentation time of frame i + one half
				DWORD ideal = (DWORD)
					(((0.5 + i) * 1000.0 * pasn->dwScale) / pasn->dwRate + 0.5);

				if (current < (pasn->dwActualLength - 1))
				{
					if (ideal >= pasn->sample_list[current + 1].pres_time)
					{
						++current;
					}
				}
				pasn->sample_map[i] = current;
			}
		}
		else
		{
			pasn->dwPerceivedLength = 1;
			pasn->sample_map = new unsigned long[1];
			if (pasn->sample_map == NULL)
			{
				mContext.mpCallbacks->SetErrorOutOfMemory();
				return;
			}
			pasn->sample_map[0] = 0;
		}
	}

	for (s = 0; s < num_streams; ++s)
	{
		pasn = streams[s];
		if (pasn->fccType == streamtypeAUDIO)
		{
			WAVEFORMATEX *wfex;
			double dur;

			// Recalculate nAvgBytesPerSec and nBlockAlign
			wfex = (WAVEFORMATEX *)pasn->pFormat;
			wfex->nAvgBytesPerSec = (DWORD)((double)pasn->i64TotalBytes / pasn->Duration + 0.5);
			wfex->nBlockAlign = (WORD)pasn->dwMaxSample;

			// Calculate dwRate and dwScale
			dur = pasn->Duration;
			pasn->dwRate = pasn->dwActualLength;
			while (pasn->dwRate < 0x3FFFFFFF && dur < 0x3FFFFFFF)
			{
				pasn->dwRate *= 2;
				dur *= 2;
			}
			pasn->dwScale = (DWORD)(dur + 0.5);

			pasn->dwPerceivedLength = pasn->dwActualLength;
			++num_audio_streams;
		}
	}
}


void VDXAPIENTRY InputFileWMV::Init(const wchar_t *szFile, IVDXInputOptions *opts)
{
	if (!mFile.Open(szFile))
	{
		mContext.mpCallbacks->SetError("Can't open file: %ls", szFile);
		return;
	}

	ParseFileRoot();
}


bool VDXAPIENTRY InputFileWMV::Append(const wchar_t *szFile)
{
	return false;
}


bool VDXAPIENTRY InputFileWMV::PromptForOptions(VDXHWND, IVDXInputOptions **ppOptions)
{
	*ppOptions = NULL;
	return false;
}


bool VDXAPIENTRY InputFileWMV::CreateOptions(const void *buf, uint32 len, IVDXInputOptions **ppOptions)
{
	*ppOptions = NULL;
	return false;
}


void VDXAPIENTRY InputFileWMV::DisplayInfo(VDXHWND hwndParent)
{
    DialogBoxParam(g_hModule, MAKEINTRESOURCE(IDD_WMV_INFO),
		(HWND)hwndParent, InfoDlgProc, (LPARAM)this);
}


bool VDXAPIENTRY InputFileWMV::GetVideoSource(int index, IVDXVideoSource **ppVS)
{
	int i;
	*ppVS = NULL;

	for (i = 0; i < num_streams; ++i)
	{
		if (streams[i]->fccType == streamtypeVIDEO)
		{
			if (index == 0)
			{
				*ppVS = new VideoSourceWMV(this, streams[i], NULL);
				if (*ppVS == NULL) break;

				(*ppVS)->AddRef();
				return true;
			}
			--index;
		}
	}

	// No video stream; choose the longest audio stream
	if (index == 0)
	{
		double max = 0.0;
		index = -1;

		for (i = 0; i < num_streams; ++i)
		{
			const ASFStreamNode& pasn = *streams[i];

			if (pasn.fccType == streamtypeAUDIO && pasn.dwRate != 0)
			{
				if (pasn.Duration > max)
				{
					max   = pasn.Duration;
					index = i;
				}
			}
		}

		if (index >= 0)
		{
			*ppVS = new VideoSourceWMV(this, NULL, streams[index]);
			if (*ppVS != NULL)
			{
				(*ppVS)->AddRef();
				return true;
			}
		}
	}

	return false;
}


bool VDXAPIENTRY InputFileWMV::GetAudioSource(int index, IVDXAudioSource **ppAS)
{
	int i;
	int idx;

	*ppAS = NULL;
	idx = index;

	for (i = 0; i < num_streams; ++i)
	{
		if (streams[i]->fccType == streamtypeAUDIO)
		{
			if (idx == 0)
			{
				*ppAS = new AudioSourceWMV(this, streams[i]);
				if (*ppAS == NULL) break;

				(*ppAS)->AddRef();
				return true;
			}
			--idx;
		}
	}

	return false;
}


///////////////////////////////////////////////////////////////////////
//
//						InputFileDriverWMV
//
///////////////////////////////////////////////////////////////////////

class InputFileDriverWMV : public vdxunknown<IVDXInputFileDriver>
{
public:
	InputFileDriverWMV(const VDXInputDriverContext& context);

	int		VDXAPIENTRY DetectBySignature(const void *pHeader, sint32 nHeaderSize, const void *pFooter, sint32 nFooterSize, sint64 nFileSize);
	bool	VDXAPIENTRY CreateInputFile(uint32 flags, IVDXInputFile **ppFile);

protected:
	const VDXInputDriverContext& mContext;
};


InputFileDriverWMV::InputFileDriverWMV(const VDXInputDriverContext& context)
	: mContext(context)
{
}


int VDXAPIENTRY InputFileDriverWMV::DetectBySignature(const void *pHeader, sint32 nHeaderSize, const void *pFooter, sint32 nFooterSize, sint64 nFileSize)
{
	return -1;
}


bool VDXAPIENTRY InputFileDriverWMV::CreateInputFile(uint32 flags, IVDXInputFile **ppFile)
{
	*ppFile = new InputFileWMV(mContext, flags);
	if (*ppFile == NULL) return false;

	(*ppFile)->AddRef();
	return true;
}


///////////////////////////////////////////////////////////////////////

bool VDXAPIENTRY wmv_create(const VDXInputDriverContext *pContext, IVDXInputFileDriver **ppDriver)
{
	*ppDriver = new InputFileDriverWMV(*pContext);
	if (*ppDriver == NULL) return false;

	(*ppDriver)->AddRef();
	return true;
}


const uint8 wmv_sig[] = {
	0x30, 255, 0x26, 255, 0xB2, 255, 0x75, 255,
	0x8E, 255, 0x66, 255, 0xCF, 255, 0x11, 255,
	0xA6, 255, 0xD9, 255, 0x00, 255, 0xAA, 255,
	0x00, 255, 0x62, 255, 0xCE, 255, 0x6C, 255
};


const VDXInputDriverDefinition wmv_input = {
	sizeof(VDXInputDriverDefinition),
	VDXInputDriverDefinition::kFlagSupportsVideo |
	VDXInputDriverDefinition::kFlagSupportsAudio |
	VDXInputDriverDefinition::kFlagNoOptions,
	0,
	sizeof(wmv_sig),
	wmv_sig,
	L"*.asf|*.wmv|*.wma",
	L"Windows Media files (*.asf,*.wmv,*.wma)|*.asf;*.wmv;*.wma",
	L"wmv",
	wmv_create
};

// (major<<24) + (minor<<16) + build.  1.4.1000 would be 0x010403E8.

//	xxxxxxxx ........ ........ ........		major
//	........ xxxxxxxx ........ ........		minor
//	........ ........ xxxxxxxx xxxxxxxx		build

const VDXPluginInfo wmv_info = {
	sizeof(VDXPluginInfo),              // size
	L"WMV plugin",                      // name
	L"fccHandler",                      // author
	L"Loads Windows Media files.",      // description
	0x02080000,                         // (major<<24) + (minor<<16) + build
	kVDXPluginType_Input,               // type
	0,                                  // flags
	10,                                 // version required
	kVDXPlugin_APIVersion,              // version used
#ifdef EXPOSE_V3
	3,                                  // version required
#else
	2,                                  // version required
#endif
	kVDXPlugin_InputDriverAPIVersion,   // version used
	&wmv_input                          // type specific info
};


///////////////////////////////////////////////////////////////////////

const VDXPluginInfo *const kPlugins[] = {
	&wmv_info,
	NULL
};

extern "C" const VDXPluginInfo *const * VDXAPIENTRY VDGetPluginInfo()
{
	return kPlugins;
}
