#ifndef _VIDEOSOURCEWMV_H_
#define _VIDEOSOURCEWMV_H_


#ifdef EXPOSE_V3
	class VideoSourceWMV : public vdxunknown<IVDXStreamSource>, public IVDXStreamSourceV3, public IVDXVideoSource {
#else
	class VideoSourceWMV : public vdxunknown<IVDXStreamSource>, public IVDXVideoSource {
#endif

public:
	VideoSourceWMV(InputFileWMV *const pp, ASFStreamNode *const vasn, ASFStreamNode *const aasn);
	~VideoSourceWMV();

	// IVDXUnknown methods
	int         VDXAPIENTRY AddRef();
	int         VDXAPIENTRY Release();
	void*       VDXAPIENTRY AsInterface(uint32 iid);

	// IVDXStreamSource methods
	void        VDXAPIENTRY GetStreamSourceInfo(VDXStreamSourceInfo&);
	bool        VDXAPIENTRY Read(sint64 lStart, uint32 lCount, void *lpBuffer, uint32 cbBuffer, uint32 *lBytesRead, uint32 *lSamplesRead);
	const void* VDXAPIENTRY GetDirectFormat();
	int         VDXAPIENTRY GetDirectFormatLen();
	ErrorMode   VDXAPIENTRY GetDecodeErrorMode();
	void        VDXAPIENTRY SetDecodeErrorMode(ErrorMode mode);
	bool        VDXAPIENTRY IsDecodeErrorModeSupported(ErrorMode mode);
	bool        VDXAPIENTRY IsVBR();
	sint64      VDXAPIENTRY TimeToPositionVBR(sint64 us);
	sint64      VDXAPIENTRY PositionToTimeVBR(sint64 samples);

#ifdef EXPOSE_V3
	// IVDXStreamSourceV3 methods
	void        VDXAPIENTRY GetStreamSourceInfoV3(VDXStreamSourceInfoV3&);
#endif

	// IVDXVideoSource methods
	void        VDXAPIENTRY GetVideoSourceInfo(VDXVideoSourceInfo& info);
	bool        VDXAPIENTRY CreateVideoDecoderModel(IVDXVideoDecoderModel **);
	bool        VDXAPIENTRY CreateVideoDecoder(IVDXVideoDecoder **);
	void        VDXAPIENTRY GetSampleInfo(sint64 sample_num, VDXVideoFrameInfo& frameInfo);
	bool        VDXAPIENTRY IsKey(sint64 sample_num);
	sint64      VDXAPIENTRY GetFrameNumberForSample(sint64 sample_num);
	sint64      VDXAPIENTRY GetSampleNumberForFrame(sint64 frame_num);
	sint64      VDXAPIENTRY GetRealFrame(sint64 frame_num);
	sint64      VDXAPIENTRY GetSampleBytePosition(sint64 sample_num);

private:
	InputFileWMV  *const parentPtr;
	ASFStreamNode *const video;
	ASFStreamNode *const audio;

	DWORD dwLength;
	DWORD dwPrevSample;
	unsigned char *frame_buffer;

	IVDXStreamSource::ErrorMode errmode;
};


#endif	// _VIDEOSOURCEWMV_H_
